@@ -112,52 +112,85 @@ def get_default_mgm():
             )
     return DEFAULT_MGM
 
-def _unsafe_split_mgm(filename):
+PROTOCOLS = [ 'root', 'srm', 'gsiftp', 'dcap' ] # Probably to be expanded
+
+def has_protocol(filename):
     """
-    Takes a properly formatted path starting with 'root:' and containing '/store'
+    Checks whether the filename contains a protocol.
+    Currently a very basic string check, which so far has been enough
     """
-    if not filename.startswith('root://'):
+    return ('://' in filename)
+
+def split_protocol_pfn(filename):
+    """
+    Splits protocol, server and logical file name from a physical file name.
+    Throws an exception of format-ensuring checks fail.
+    """
+    if not has_protocol(filename):
         raise ValueError(
-            'Cannot split mgm; passed filename: {0}'
+            'Attempted to get protocol from {0}, but there'
+            ' does not seem to be any.'
             .format(filename)
             )
-    elif not '/store' in filename:
+    protocol, rest = filename.split('://',1)
+    if not '//' in rest:
         raise ValueError(
-            'No substring \'/store\' in filename {0}'
+            'Could not determine server and logical file name from {0}'
             .format(filename)
             )
-    i = filename.index('/store')
-    mgm = filename[:i]
-    lfn = filename[i:]
-    return mgm, lfn
+    server, lfn = rest.split('//',1)
+    lfn = '/' + lfn # Restore the opening slash that was dropped in the split
+    return protocol, server, lfn
+
+def _split_mgm_pfn(filename):
+    """
+    Splits mgm and logical file name from a physical file name.
+    Throws an exception of format-ensuring checks fail.
+    """
+    protocol, server, lfn = split_protocol_pfn(filename)
+    return protocol + '://' + server, lfn
+
+def _join_protocol_server_lfn(protocol, server, lfn):
+    """
+    Joins protocol, server and lfn into a physical filename.
+    Ensures formatting to some extent.
+    """
+    protocol = protocol.replace(':', '') # Remove any ':' from the protocol
+    server = server.strip('/') # Strip trailing or opening slashes
+    if not lfn.startswith('/'):
+        raise ValueError(
+            'Logical file name {0} does not seem to be formatted correctly'
+            .format(lfn)
+            )
+    return protocol + '://' + server + '/' + lfn
 
 def split_mgm(path, mgm=None):
     """
     Returns the mgm and lfn that the user most likely intended to
-    if path starts with 'root://', the mgm is taken from the path
+    if path has a protocol string (e.g. 'root://...'), the mgm is taken from the path
     if mgm is passed, it is used as is
     if mgm is passed AND the path starts with 'root://' AND the mgm's don't agree,
       an exception is thrown
     if mgm is None and path has no mgm, the default variable DEFAULT_MGM is taken
     """
-    if path.startswith('root://'):
-        _mgm, lfn = _unsafe_split_mgm(path)
-        if not(mgm is None) and not _mgm == mgm:
+    if has_protocol(path):
+        mgm_from_path, lfn = _split_mgm_pfn(path)
+        if not(mgm is None) and not mgm_from_path == mgm:
             raise ValueError(
                 'Conflicting mgms determined from path and passed argument: '
                 'From path {0}: {1}, from argument: {2}'
-                .format(path, _mgm, mgm)
+                .format(path, mgm_from_path, mgm)
                 )
-        mgm = _mgm
+        mgm = mgm_from_path
     elif mgm is None:
         mgm = get_default_mgm()
         lfn = path
     else:
         lfn = path
     # Sanity check
-    if not lfn.startswith('/store'):
+    if not lfn.startswith('/'):
         raise ValueError(
-            'LFN {0} does not start with \'/store\'; something is wrong'
+            'LFN {0} does not start with \'/\'; something is wrong'
             .format(lfn)
             )
     return mgm, lfn
@@ -165,11 +198,11 @@ def split_mgm(path, mgm=None):
 def _join_mgm_lfn(mgm, lfn):
     """
     Joins mgm and lfn, ensures correct formatting.
-    Will throw an exception of the lfn does not start with '/store'
+    Will throw an exception of the lfn does not start with '/'
     """
-    if not lfn.startswith('/store'):
+    if not lfn.startswith('/'):
         raise ValueError(
-            'This function expects filenames that start with \'/store\''
+            'This function expects filenames that start with \'/\''
             )
     if not mgm.endswith('/'): mgm += '/'
     return mgm + lfn
@@ -179,13 +212,20 @@ def format(path, mgm=None):
     Formats a path to ensure it is a path on the SE.
     Can take:
     - Just path starting with 'root:' - nothing really happens
-    - Just path starting with '/store' - the default mgm is used
+    - Just path starting with '/' - the default mgm is used
     - Path starting with 'root:' and an mgm - an exception is thrown in case of conflict
-    - Path starting with '/store' and an mgm - mgm and path are joined
+    - Path starting with '/' and an mgm - mgm and path are joined
     """
     mgm, lfn = split_mgm(path, mgm=mgm)
     return _join_mgm_lfn(mgm, lfn)
 
+def get_protocol(path, mgm=None):
+    """
+    Returns the protocol contained in the path string
+    """
+    path = format(path, mgm)
+    return path.split('://')[0]
+
 # _______________________________________________________
 # Interactions with SE
 
@@ -250,53 +290,120 @@ def is_file_or_dir(path):
             .format(' '.join(cmd), status)
             )
 
-def cp(src, dst, create_parent_directory=True):
+def cp(src, dst, method='auto', **kwargs):
     """
     Copies a file `src` to the storage element.
     Does not format `src` or `dst`; user is responsible for formatting.
+
+    The method can be 'auto', 'xrdcp', or 'gfal-copy'. If 'auto', a heuristic
+    will be applied to determine whether to best use xrdcp or gfal-copy.
     """
     logger.warning('Copying %s --> %s', src, dst)
-    if create_parent_directory:
-        cmd = [ 'xrdcp', '-s', '-p', src, dst ]
-    else:
-        cmd = [ 'xrdcp', '-s', src, dst ]
+    methods = {
+        'xrdcp' : _cp_xrdcp,
+        'gfal-copy' : _cp_gfal,
+        }
+    # Heuristic to determine what copy method to use
+    if method == 'auto':
+        for file in [src, dst]:
+            if has_protocol(file):
+                protocol = get_protocol(file)
+                if protocol == 'root':
+                    method = 'xrdcp'
+                else:
+                    method = 'gfal-copy'
+                break
+        else:
+            logger.debug(
+                'No protocols specified in either src ({0}) or dst ({1}); using xrdcp'
+                .format(src, dst)
+                )
+            method = 'xrdcp'
+    # Execute the copy method
+    try:
+        methods[method](src, dst, **kwargs)
+    except KeyError:
+        logger.error('Method %s is not a valid copying method!', method)
+        raise
+
+def _cp_xrdcp(src, dst, create_parent_directory=True, verbose=True):
+    cmd = [ 'xrdcp', src, dst ]
+    if not verbose: cmd.insert(1, '-s')
+    if create_parent_directory: cmd.insert(1, '-p')
     run_command(cmd)
 
-def cp_to_se(src, dst, create_parent_directory=True):
+def _cp_gfal(src, dst, create_parent_directory=True, verbose=True):
+    cmd = [ 'gfal-copy', '-t', '180', src, dst ]
+    if create_parent_directory: cmd.insert(1, '-p')
+    if verbose: cmd.insert(1, '-v')
+    run_command(cmd)
+
+def cp_to_se(src, dst, **kwargs):
     """
     Like cp, but assumes dst is a location on a storage element and src is local
     """
-    cp(src, format(dst))
+    cp(src, format(dst), **kwargs)
 
-def cp_from_se(src, dst, create_parent_directory=True):
+def cp_from_se(src, dst, **kwargs):
     """
     Like cp, but assumes src is a location on a storage element and dst is local
     """
-    cp(format(src), dst)
+    cp(format(src), dst, **kwargs)
 
 class Inode(object):
     """
     Basic container of information returned by xrdfs MGM ls -l PATH:
     permission string, modification time, size, and path
     """
-    def __init__(self, statline, mgm=None):
+
+    @classmethod
+    def from_path(cls, path, mgm=None):
+        import datetime
+        path = format(path, mgm) # Ensures consistency in input
+        mgm, path = split_mgm(path)
+        size = None
+        modtime = None
+        isdir = None
+        cmd = [ 'xrdfs', mgm, 'stat', path ]
+        output = run_command(cmd)
+        for l in output:
+            l = l.strip()
+            if l.startswith('Size:'):
+                size = int(l.split()[1])
+            elif l.startswith('MTime:'):
+                timestamp = l.replace('MTime:', '').strip()
+                modtime = datetime.datetime.strptime(timestamp, '%Y-%m-%d %H:%M:%S')
+            elif l.startswith('Flags:'):
+                isdir = 'IsDir' in l
+        if size is None: raise RuntimeError('Could not extract size from stat:\n{0}'.format(output))
+        if modtime is None: raise RuntimeError('Could not extract modtime from stat:\n{0}'.format(output))
+        if isdir is None: raise RuntimeError('Could not extract isdir from stat:\n{0}'.format(output))
+        return cls(path, modtime, isdir, size)
+
+    @classmethod
+    def from_statline(cls, statline, mgm):
+        import datetime
         components = statline.strip().split()
         if not len(components) == 5:
             raise RuntimeError(
                 'Expected 5 components for stat line:\n{0}'
                 .format(statline)
                 )
-        self.permissions = components[0]
-        self.isdir = self.permissions.startswith('d')
+        isdir = components[0].startswith('d')
+        modtime = datetime.datetime.strptime(components[1] + ' ' + components[2], '%Y-%m-%d %H:%M:%S')
+        size = int(components[3])
+        path = format(components[4], mgm)
+        return cls(path, modtime, isdir, size)
+
+    def __init__(self, path, modtime, isdir, size):
+        self.path = path
+        self.modtime = modtime
+        self.isdir = isdir
+        self.size = size
+        # Some derived properties
         self.isfile = not(self.isdir)
-        import datetime
-        self.modtime = datetime.datetime.strptime(components[1] + ' ' + components[2], '%Y-%m-%d %H:%M:%S')
-        self.size = components[3]
         self.size_human = bytes_to_human_readable(float(self.size))
-        self.localpath = components[4]
-        if mgm:
-            self.mgm = mgm
-            self.path = format(self.localpath, mgm)
+        self.basename = osp.basename(path)
 
     def __repr__(self):
         if len(self.path) > 40:
@@ -305,7 +412,7 @@ class Inode(object):
             shortpath = self.path
         return super(Inode, self).__repr__().replace('object', shortpath)
 
-def ls(path, stat=False, assume_directory=False):
+def ls(path, stat=False, assume_directory=False, no_expand_directory=False):
     """
     Lists all files and directories in a directory on the SE.
     It first checks whether the path exists and is a file or a directory.
@@ -314,8 +421,13 @@ def ls(path, stat=False, assume_directory=False):
     If it is a directory, it returns a list of the directory contents (formatted)
 
     If stat is True, it returns Inode objects which contain more information beyond just the path
+
     If assume_directory is True, the first check is not performed and the algorithm assumes
-    the user took care to pass a path to a directory.
+    the user took care to pass a path to a directory. This saves a request to the SE, which might
+    matter in the walk algorithm. For singular use, assume_directory should be set to False.
+
+    If no_expand_directory is True, the contents of the directory are not listed, and instead
+    a formatted path to the directory is returned (similar to unix's ls -d)
     """
     mgm, path = split_mgm(path)
     if assume_directory:
@@ -326,22 +438,27 @@ def ls(path, stat=False, assume_directory=False):
     if status == 0:
         raise RuntimeError('Path \'{0}\' does not exist'.format(path))
     elif status == 1:
-        # It's a directory; return contents
-        cmd = [ 'xrdfs', mgm, 'ls', path ]
-        if stat: cmd.append('-l')
-        output = run_command(cmd)
-        contents = []
-        for l in output:
-            l = l.strip()
-            if not len(l): continue
-            if stat:
-                contents.append(Inode(l, mgm))
-            else:
-                contents.append(format(l, mgm))
-        return contents
+        # It's a directory
+        if no_expand_directory:
+            # If not expanding, just return a formatted path to the directory
+            return [Inode.from_path(path, mgm)] if stat else [_join_mgm_lfn(mgm, path)]
+        else:
+            # List the contents of the directory
+            cmd = [ 'xrdfs', mgm, 'ls', path ]
+            if stat: cmd.append('-l')
+            output = run_command(cmd)
+            contents = []
+            for l in output:
+                l = l.strip()
+                if not len(l): continue
+                if stat:
+                    contents.append(Inode.from_statline(l, mgm))
+                else:
+                    contents.append(format(l, mgm))
+            return contents
     elif status == 2:
         # It's a file; just return the path to the file
-        return [_join_mgm_lfn(mgm, path)]
+        return [Inode.from_path(path, mgm)] if stat else [_join_mgm_lfn(mgm, path)]
 
 MAX_RECURSION_DEPTH = 20
 
@@ -390,9 +507,9 @@ def _walk(path, stat, counter):
     contents = ls(path, stat=True, assume_directory=True)
     counter.plus_one()
     files = [ c for c in contents if c.isfile ]
-    files.sort(key=lambda f: f.localpath)
+    files.sort(key=lambda f: f.basename)
     directories = [ c for c in contents if c.isdir ]
-    directories.sort(key=lambda d: d.localpath)
+    directories.sort(key=lambda d: d.basename)
     if stat:
         yield path, directories, files
     else:
@@ -449,13 +566,14 @@ def ls_root(paths):
 def ls_wildcard(pattern, stat=False):
     """
     Like ls, but accepts wildcards * .
+    Directories are *not* expanded.
 
     The algorithm is like `walk`, but discards directories that don't fit the pattern
     early.
     Still the number of requests can grow quickly; a limited number of wildcards is advised.
     """
     pattern = format(pattern)
-    if not '*' in pattern: return ls(pattern, stat=stat)
+    if not '*' in pattern: return ls(pattern, stat=stat, no_expand_directory=True)
     pattern_level = pattern.count('/')
     logger.debug('Level is %s for path %s', pattern_level, pattern)
     import re
@@ -504,6 +622,13 @@ def _hadd(root_files, dst, dry=False):
     finally:
         debug(False)
 
+def _hadd_packed(tup):
+    """
+    Just unpacks an input tuple and calls _hadd.
+    Needed to work with multiprocessing.
+    """
+    return _hadd(*tup)
+
 def hadd(src, dst, dry=False):
     """
     Calls `ls_root` on `src` in order to be able to pass directories, then hadds.
@@ -548,17 +673,20 @@ def _hadd_chunks(root_files, dst, n_threads=6, chunk_size=200, tmpdir='/tmp', dr
     try:
         debug(True)
         chunk_rootfiles = []
-        p = mp.Pool(n_threads)
+        # First compile list of function arguments
+        func_args = []
         for i_chunk in range(n_chunks):
             chunk = root_files[ i_chunk*chunk_size : (i_chunk+1)*chunk_size ]
             chunk_dst = osp.join(tmpdir, 'chunk{0}.root'.format(i_chunk))
-            if dry:
-                logger.debug('hadding %s --> %s', ' '.join(chunk), chunk_dst)
-            else:
-                p.map(_hadd, (chunk, chunk_dst, dry))
+            func_args.append([chunk, chunk_dst, dry])
             chunk_rootfiles.append(chunk_dst)
-        p.close()
-        p.join()
+            if dry: logger.debug('hadding %s --> %s', ' '.join(chunk), chunk_dst)
+        # Submit to multiprocessing in one go:
+        if not dry:
+            p = mp.Pool(n_threads)
+            p.map(_hadd_packed, func_args)
+            p.close()
+            p.join()
         # Merge the chunks into the final destination, potentially with another chunked merge
         _hadd_chunks(chunk_rootfiles, dst, n_threads, chunk_size, tmpdir, dry)
 
@@ -591,14 +719,14 @@ def cli_detect_fnal():
     return mgm
 
 def cli_flexible_format(lfn, mgm=None):
-    if not lfn.startswith('root:') and not lfn.startswith('/'):
+    if has_protocol(lfn) and not lfn.startswith('/'):
         try:
             prefix = '/store/user/' + os.environ['USER']
             logger.warning('Pre-fixing %s', prefix)
             lfn = os.path.join(prefix, lfn)
         except KeyError:
             pass
-    if lfn.startswith('root:'):
+    if has_protocol(lfn):
         return format(lfn)
     else:
         return format(lfn, mgm)
